package org.xbib.jdbc.pool;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.LongAccumulator;
import java.util.concurrent.atomic.LongAdder;
import java.util.function.Function;
import javax.sql.XAConnection;
import org.xbib.jdbc.pool.MetricsRepository.EmptyMetricsRepository;
import org.xbib.jdbc.pool.api.XbibDataSource;
import org.xbib.jdbc.pool.api.XbibDataSourceListener;
import org.xbib.jdbc.pool.api.XbibPoolInterceptor;
import org.xbib.jdbc.pool.api.cache.ConnectionCache;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration;
import org.xbib.jdbc.pool.api.transaction.TransactionIntegration;
import org.xbib.jdbc.pool.util.PriorityScheduledExecutor;
import org.xbib.jdbc.pool.util.StampedCopyOnWriteArrayList;
import org.xbib.jdbc.pool.util.XbibSynchronizer;
import static java.lang.Integer.toHexString;
import static java.lang.System.identityHashCode;
import static java.lang.System.nanoTime;
import static java.lang.Thread.currentThread;
import static java.util.Collections.unmodifiableList;
import static java.util.concurrent.TimeUnit.NANOSECONDS;
import static java.util.stream.Collectors.joining;
import static java.util.stream.Collectors.toList;
import static org.xbib.jdbc.pool.ConnectionHandler.State.CHECKED_IN;
import static org.xbib.jdbc.pool.ConnectionHandler.State.CHECKED_OUT;
import static org.xbib.jdbc.pool.ConnectionHandler.State.FLUSH;
import static org.xbib.jdbc.pool.ConnectionHandler.State.VALIDATION;
import static org.xbib.jdbc.pool.api.XbibDataSource.FlushMode.GRACEFUL;
import static org.xbib.jdbc.pool.api.XbibDataSource.FlushMode.LEAK;
import static org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.MultipleAcquisitionAction.OFF;
import static org.xbib.jdbc.pool.util.InterceptorHelper.fireOnConnectionAcquiredInterceptor;
import static org.xbib.jdbc.pool.util.InterceptorHelper.fireOnConnectionCreateInterceptor;
import static org.xbib.jdbc.pool.util.InterceptorHelper.fireOnConnectionDestroyInterceptor;
import static org.xbib.jdbc.pool.util.InterceptorHelper.fireOnConnectionReturnInterceptor;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionAcquire;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionCreation;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionDestroy;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionFlush;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionLeak;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionReap;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionReturn;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionValidation;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionAcquired;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionCreation;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionDestroy;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionFlush;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionInvalid;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionLeak;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionPooled;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionReap;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionReturn;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionValid;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnInfo;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnWarning;

public final class ConnectionPool implements Pool {

    private static final AtomicInteger HOUSEKEEP_COUNT = new AtomicInteger();

    private final XbibConnectionPoolConfiguration configuration;
    private final XbibDataSourceListener[] listeners;

    private final StampedCopyOnWriteArrayList<ConnectionHandler> allConnections;

    private final XbibSynchronizer synchronizer;
    private final ConnectionFactory connectionFactory;
    private final PriorityScheduledExecutor housekeepingExecutor;
    private final TransactionIntegration transactionIntegration;

    private final boolean borrowValidationEnabled;
    private final boolean idleValidationEnabled;
    private final boolean leakEnabled;
    private final boolean validationEnabled;
    private final boolean reapEnabled;

    private final LongAccumulator maxUsed = new LongAccumulator(Math::max, Long.MIN_VALUE);
    private final LongAdder activeCount = new LongAdder();
    private final ConnectionCache localCache;
    private MetricsRepository metricsRepository;
    private List<XbibPoolInterceptor> interceptors;

    public ConnectionPool(XbibConnectionPoolConfiguration configuration, XbibDataSourceListener... listeners) {
        this.configuration = configuration;
        this.listeners = listeners;

        allConnections = new StampedCopyOnWriteArrayList<>(ConnectionHandler.class);
        localCache = configuration.connectionCache();

        synchronizer = new XbibSynchronizer();
        connectionFactory = new ConnectionFactory(configuration.connectionFactoryConfiguration(), listeners);
        housekeepingExecutor = new PriorityScheduledExecutor(1, "xbib-" + HOUSEKEEP_COUNT.incrementAndGet(), listeners);
        transactionIntegration = configuration.transactionIntegration();

        borrowValidationEnabled = configuration.validateOnBorrow();
        idleValidationEnabled = !configuration.validateOnBorrow() && !configuration.idleValidationTimeout().isZero();
        leakEnabled = !configuration.leakTimeout().isZero();
        validationEnabled = !configuration.validationTimeout().isZero();
        reapEnabled = !configuration.reapTimeout().isZero();
    }

    public void init() {
        if (configuration.acquisitionTimeout().compareTo(configuration.connectionFactoryConfiguration().loginTimeout()) < 0) {
            fireOnWarning(listeners, "Login timeout should be smaller than acquisition timeout");
        }

        if (leakEnabled) {
            housekeepingExecutor.schedule(new LeakTask(), configuration.leakTimeout().toNanos(), NANOSECONDS);
        }
        if (validationEnabled) {
            housekeepingExecutor.schedule(new ValidationTask(), configuration.validationTimeout().toNanos(), NANOSECONDS);
        }
        if (reapEnabled) {
            housekeepingExecutor.schedule(new ReapTask(), configuration.reapTimeout().toNanos(), NANOSECONDS);
        }
        transactionIntegration.addResourceRecoveryFactory(connectionFactory.hasRecoveryCredentials() ? connectionFactory : this);

        // fill to the initial size
        if (configuration.initialSize() < configuration.minSize()) {
            fireOnInfo(listeners, "Initial size smaller than min. Connections will be created when necessary");
        } else if (configuration.initialSize() > configuration.maxSize()) {
            fireOnInfo(listeners, "Initial size bigger than max. Connections will be destroyed as soon as they return to the pool");
        }
        for (int n = configuration.initialSize(); n > 0; n--) {
            housekeepingExecutor.executeNow(new CreateConnectionTask().initial());
        }
    }

    public XbibConnectionPoolConfiguration getConfiguration() {
        return configuration;
    }

    public int defaultJdbcIsolationLevel() {
        return connectionFactory.defaultJdbcIsolationLevel();
    }

    public XbibDataSourceListener[] getListeners() {
        return listeners;
    }

    public List<XbibPoolInterceptor> getPoolInterceptors() {
        return unmodifiableList(interceptors);
    }

    public void setPoolInterceptors(Collection<? extends XbibPoolInterceptor> list) {
        if (list.stream().anyMatch(i -> i.getPriority() < 0)) {
            throw new IllegalArgumentException("Negative priority values on XbibPoolInterceptor are reserved.");
        }
        if (list.isEmpty() && (interceptors == null || interceptors.isEmpty())) {
            return;
        }

        interceptors = list.stream().sorted(XbibPoolInterceptor.DEFAULT_COMPARATOR).collect(toList());

        Function<XbibPoolInterceptor, String> interceptorName = i -> i.getClass().getName() + "@" + toHexString(identityHashCode(i)) + " (priority " + i.getPriority() + ")";
        fireOnInfo(listeners, "Pool interceptors: " + interceptors.stream().map(interceptorName).collect(joining(" >>> ", "[", "]")));
    }

    public void flushPool(XbibDataSource.FlushMode mode) {
        if (mode == LEAK && !leakEnabled) {
            fireOnWarning(listeners, "Flushing leak connections with no specified leak timout.");
            return;
        }
        housekeepingExecutor.execute(new FlushTask(mode));
    }

    @Override
    public void close() {
        transactionIntegration.removeResourceRecoveryFactory(connectionFactory.hasRecoveryCredentials() ? connectionFactory : this);

        for (Runnable task : housekeepingExecutor.shutdownNow()) {
            if (task instanceof DestroyConnectionTask) {
                task.run();
            }
        }

        for (ConnectionHandler handler : allConnections) {
            handler.setState(FLUSH);
            new DestroyConnectionTask(handler).run();
        }
        allConnections.clear();
        activeCount.reset();

        synchronizer.release(synchronizer.getQueueLength());
    }

    @Override
    public boolean isRecoverable() {
        return connectionFactory.isRecoverable();
    }

    @Override
    public XAConnection getRecoveryConnection() throws SQLException {
        long stamp = beforeAcquire();
        checkMultipleAcquisition();
        ConnectionHandler checkedOutHandler = null;

        try {
            do {
                checkedOutHandler = (ConnectionHandler) localCache.get();
                if (checkedOutHandler == null) {
                    checkedOutHandler = handlerFromSharedCache();
                }
            } while ((borrowValidationEnabled && !borrowValidation(checkedOutHandler))
                    || (idleValidationEnabled && !idleValidation(checkedOutHandler)));

            activeCount.increment();
            fireOnConnectionAcquiredInterceptor(interceptors, checkedOutHandler);
            afterAcquire(stamp, checkedOutHandler, false);
            return checkedOutHandler.xaConnectionWrapper();
        } catch (Throwable t) {
            if (checkedOutHandler != null) {
                checkedOutHandler.setState(CHECKED_OUT, CHECKED_IN);
            }
            throw t;
        }
    }

    private long beforeAcquire() throws SQLException {
        fireBeforeConnectionAcquire(listeners);
        if (housekeepingExecutor.isShutdown()) {
            throw new SQLException("This pool is closed and does not handle any more connections!");
        }
        return metricsRepository.beforeConnectionAcquire();
    }

    private void checkMultipleAcquisition() throws SQLException {
        if (configuration.multipleAcquisition() != OFF) {
            for (ConnectionHandler handler : allConnections) {
                if (handler.getHoldingThread() == currentThread()) {
                    switch (configuration.multipleAcquisition()) {
                        case STRICT:
                            throw new SQLException("Acquisition of multiple connections by the same Thread.");
                        case WARN:
                            fireOnWarning(listeners, "Acquisition of multiple connections by the same Thread. This can lead to pool exhaustion and eventually a deadlock!");
                    }
                    break;
                }
            }
        }
    }

    @Override
    public Connection getConnection() throws SQLException {
        long stamp = beforeAcquire();

        ConnectionHandler checkedOutHandler = handlerFromTransaction();
        if (checkedOutHandler != null) {
            // AG-140 - If associate throws here is fine, it's assumed the synchronization that returns the connection has been registered
            transactionIntegration.associate(checkedOutHandler, checkedOutHandler.getXaResource());
            afterAcquire(stamp, checkedOutHandler, true);
            return checkedOutHandler.connectionWrapper();
        }
        checkMultipleAcquisition();

        try {
            do {
                checkedOutHandler = (ConnectionHandler) localCache.get();
                if (checkedOutHandler == null) {
                    checkedOutHandler = handlerFromSharedCache();
                }
            } while ((borrowValidationEnabled && !borrowValidation(checkedOutHandler))
                    || (idleValidationEnabled && !idleValidation(checkedOutHandler)));
            transactionIntegration.associate(checkedOutHandler, checkedOutHandler.getXaResource());

            activeCount.increment();
            fireOnConnectionAcquiredInterceptor(interceptors, checkedOutHandler);
            afterAcquire(stamp, checkedOutHandler, true);
            return checkedOutHandler.connectionWrapper();
        } catch (Throwable t) {
            if (checkedOutHandler != null) {
                // AG-140 - Return the connection to the pool to prevent leak
                checkedOutHandler.setState(CHECKED_OUT, CHECKED_IN);
            }
            throw t;
        }
    }

    private ConnectionHandler handlerFromTransaction() throws SQLException {
        return (ConnectionHandler) transactionIntegration.getTransactionAware();
    }

    private ConnectionHandler handlerFromSharedCache() throws SQLException {
        long remaining = configuration.acquisitionTimeout().toNanos();
        remaining = remaining > 0 ? remaining : Long.MAX_VALUE;
        Future<ConnectionHandler> task = null;
        try {
            for (; ; ) {
                // If min-size increases, create a connection right away
                if (allConnections.size() < configuration.minSize()) {
                    task = housekeepingExecutor.executeNow(new CreateConnectionTask());
                }
                // Try to find an available connection in the pool
                for (ConnectionHandler handler : allConnections) {
                    if (handler.acquire()) {
                        return handler;
                    }
                }
                // If no connections are available and there is room, create one
                if (task == null && allConnections.size() < configuration.maxSize()) {
                    task = housekeepingExecutor.executeNow(new CreateConnectionTask());
                }
                long start = nanoTime();
                if (task == null) {
                    // Pool full, will have to wait for a connection to be returned
                    if (!synchronizer.tryAcquireNanos(synchronizer.getStamp(), remaining)) {
                        throw new SQLException("Sorry, acquisition timeout!");
                    }
                } else {
                    // Wait for the new connection instead of the synchronizer to propagate any exception on connection establishment
                    ConnectionHandler handler = task.get(remaining, NANOSECONDS);
                    if (handler != null && handler.acquire()) {
                        return handler;
                    }
                    task = null;
                }
                remaining -= nanoTime() - start;
            }
        } catch (InterruptedException e) {
            currentThread().interrupt();
            throw new SQLException("Interrupted while acquiring");
        } catch (ExecutionException e) {
            throw unwrapExecutionException(e);
        } catch (RejectedExecutionException | CancellationException e) {
            throw new SQLException("Can't create new connection as the pool is shutting down", e);
        } catch (TimeoutException e) {
            task.cancel(true);
            // AG-201: Last effort. Connections may have returned to the pool while waiting.
            for (ConnectionHandler handler : allConnections) {
                if (handler.acquire()) {
                    return handler;
                }
            }
            throw new SQLException("Acquisition timeout while waiting for new connection", e);
        }
    }

    private SQLException unwrapExecutionException(ExecutionException ee) {
        try {
            throw ee.getCause();
        } catch (RuntimeException | Error re) {
            throw re;
        } catch (SQLException se) {
            return se;
        } catch (Throwable t) {
            return new SQLException("Exception while creating new connection", t);
        }
    }

    private boolean idleValidation(ConnectionHandler handler) {
        if (!handler.isIdle(configuration.idleValidationTimeout())) {
            return true;
        }
        return borrowValidation(handler);
    }

    private boolean borrowValidation(ConnectionHandler handler) {
        if (handler.setState(CHECKED_OUT, VALIDATION)) {
            return performValidation(handler, CHECKED_OUT);
        }
        return false;
    }

    private boolean performValidation(ConnectionHandler handler, ConnectionHandler.State targetState) {
        fireBeforeConnectionValidation(listeners, handler);
        if (handler.isValid() && handler.setState(VALIDATION, targetState)) {
            fireOnConnectionValid(listeners, handler);
            synchronizer.releaseConditional();
            return true;
        } else {
            removeFromPool(handler);
            metricsRepository.afterConnectionInvalid();
            fireOnConnectionInvalid(listeners, handler);
            return false;
        }
    }

    private void afterAcquire(long metricsStamp, ConnectionHandler checkedOutHandler, boolean verifyEnlistment) throws SQLException {
        metricsRepository.afterConnectionAcquire(metricsStamp);
        fireOnConnectionAcquired(listeners, checkedOutHandler);

        if (verifyEnlistment && !checkedOutHandler.isEnlisted()) {
            switch (configuration.transactionRequirement()) {
                case STRICT:
                    returnConnectionHandler(checkedOutHandler);
                    throw new SQLException("Connection acquired without transaction.");
                case WARN:
                    fireOnWarning(listeners, new SQLException("Connection acquired without transaction."));
            }
        }
        if (leakEnabled || reapEnabled) {
            checkedOutHandler.touch();
        }
        if (leakEnabled || configuration.multipleAcquisition() != OFF) {
            if (checkedOutHandler.getHoldingThread() != null && checkedOutHandler.getHoldingThread() != currentThread()) {
                Throwable warn = new Throwable("Shared connection between threads '" + checkedOutHandler.getHoldingThread().getName() + "' and '" + currentThread().getName() + "'");
                warn.setStackTrace(checkedOutHandler.getHoldingThread().getStackTrace());
                fireOnWarning(listeners, warn);
            }
            checkedOutHandler.setHoldingThread(currentThread());
            if (configuration.enhancedLeakReport()) {
                checkedOutHandler.setAcquisitionStackTrace(currentThread().getStackTrace());
            }
        }
    }

    // --- //

    public void returnConnectionHandler(ConnectionHandler handler) throws SQLException {
        fireBeforeConnectionReturn(listeners, handler);
        if (leakEnabled) {
            handler.setHoldingThread(null);
            if (configuration.enhancedLeakReport()) {
                handler.setAcquisitionStackTrace(null);
            }
        }
        if (idleValidationEnabled || reapEnabled) {
            handler.touch();
        }
        try {
            if (!transactionIntegration.disassociate(handler)) {
                return;
            }
        } catch (Throwable ignored) {
        }

        activeCount.decrement();

        // resize on change of max-size, or flush on close
        int currentSize = allConnections.size();
        if ((currentSize > configuration.maxSize() && currentSize > configuration.minSize()) || configuration.flushOnClose()) {
            handler.setState(FLUSH);
            removeFromPool(handler);
            metricsRepository.afterConnectionReap();
            fireOnConnectionReap(listeners, handler);
            return;
        }

        try {
            handler.resetConnection();
        } catch (SQLException sqlException) {
            fireOnWarning(listeners, sqlException);
        }
        localCache.put(handler);
        fireOnConnectionReturnInterceptor(interceptors, handler);

        if (handler.setState(CHECKED_OUT, CHECKED_IN)) {
            // here the handler is already up for grabs
            synchronizer.releaseConditional();
            metricsRepository.afterConnectionReturn();
            fireOnConnectionReturn(listeners, handler);
        } else {
            // handler not in CHECKED_OUT implies FLUSH
            removeFromPool(handler);
            metricsRepository.afterConnectionFlush();
            fireOnConnectionFlush(listeners, handler);
        }
    }

    private void removeFromPool(ConnectionHandler handler) {
        allConnections.remove(handler);
        synchronizer.releaseConditional();
        housekeepingExecutor.execute(new FillTask());
        housekeepingExecutor.execute(new DestroyConnectionTask(handler));
    }

    // --- Exposed statistics //

    @Override
    public void onMetricsEnabled(boolean metricsEnabled) {
        metricsRepository = metricsEnabled ? new DefaultMetricsRepository(this) : new EmptyMetricsRepository();
    }

    public MetricsRepository getMetrics() {
        return metricsRepository;
    }

    public long activeCount() {
        return activeCount.sum();
    }

    public long availableCount() {
        return allConnections.size() - activeCount.sum();
    }

    public long maxUsedCount() {
        return maxUsed.get();
    }

    public void resetMaxUsedCount() {
        maxUsed.reset();
    }

    public long awaitingCount() {
        return synchronizer.getQueueLength();
    }

    // --- health check //

    @Override
    public boolean isHealthy(boolean newConnection) throws SQLException {
        ConnectionHandler healthHandler;
        Future<ConnectionHandler> task = null;
        if (newConnection) {
            try {
                do {
                    task = housekeepingExecutor.executeNow(new CreateConnectionTask().initial());
                    healthHandler = task.get(configuration.acquisitionTimeout().isZero() ? Long.MAX_VALUE : configuration.acquisitionTimeout().toNanos(), NANOSECONDS);
                } while (!healthHandler.setState(CHECKED_IN, VALIDATION));
            } catch (InterruptedException e) {
                currentThread().interrupt();
                throw new SQLException("Interrupted while acquiring");
            } catch (ExecutionException ee) {
                throw unwrapExecutionException(ee);
            } catch (RejectedExecutionException | CancellationException e) {
                throw new SQLException("Can't create new connection as the pool is shutting down", e);
            } catch (TimeoutException e) {
                task.cancel(true);
                throw new SQLException("Acquisition timeout on health check");
            }
        } else {
            healthHandler = handlerFromSharedCache();
            healthHandler.setState(CHECKED_OUT, VALIDATION);
        }
        return performValidation(healthHandler, CHECKED_IN);
    }

    // --- create //

    private final class CreateConnectionTask implements Callable<ConnectionHandler> {

        private boolean initial;

        // initial connections do not take configuration.maxSize into account
        private CreateConnectionTask initial() {
            initial = true;
            return this;
        }

        @Override
        public ConnectionHandler call() throws SQLException {
            if (!initial && allConnections.size() >= configuration.maxSize()) {
                return null;
            }
            fireBeforeConnectionCreation(listeners);
            long metricsStamp = metricsRepository.beforeConnectionCreation();

            try {
                ConnectionHandler handler = new ConnectionHandler(connectionFactory.createConnection(), ConnectionPool.this);
                metricsRepository.afterConnectionCreation(metricsStamp);

                if (!configuration.maxLifetime().isZero()) {
                    handler.setMaxLifetimeTask(housekeepingExecutor.schedule(new FlushTask(GRACEFUL, handler), configuration.maxLifetime().toNanos(), NANOSECONDS));
                }

                fireOnConnectionCreation(listeners, handler);
                fireOnConnectionCreateInterceptor(interceptors, handler);

                handler.setState(CHECKED_IN);
                allConnections.add(handler);

                maxUsed.accumulate(allConnections.size());
                fireOnConnectionPooled(listeners, handler);

                return handler;
            } catch (SQLException e) {
                fireOnWarning(listeners, e);
                throw e;
            } catch (Throwable t) {
                fireOnWarning(listeners, "Failed to create connection due to " + t.getClass().getSimpleName());
                throw t;
            } finally {
                // not strictly needed, but not harmful either
                synchronizer.releaseConditional();
            }
        }
    }

    // --- flush //

    private final class FlushTask implements Runnable {

        private final XbibDataSource.FlushMode mode;
        private final ConnectionHandler handler;

        @SuppressWarnings("WeakerAccess")
        FlushTask(XbibDataSource.FlushMode mode) {
            this.mode = mode;
            this.handler = null;
        }

        @SuppressWarnings({"WeakerAccess", "SameParameterValue"})
        FlushTask(XbibDataSource.FlushMode mode, ConnectionHandler handler) {
            this.mode = mode;
            this.handler = handler;
        }

        @Override
        public void run() {
            for (ConnectionHandler ch : handler != null ? Collections.singleton(handler) : allConnections) {
                fireBeforeConnectionFlush(listeners, ch);
                flush(mode, ch);
            }
            afterFlush(mode);
        }

        private void flush(XbibDataSource.FlushMode mode, ConnectionHandler handler) {
            switch (mode) {
                case ALL:
                    handler.setState(FLUSH);
                    flushHandler(handler);
                    break;
                case GRACEFUL:
                    if (handler.setState(CHECKED_IN, FLUSH)) {
                        flushHandler(handler);
                    } else if (!handler.setState(CHECKED_OUT, FLUSH) && handler.isAcquirable()) {
                        // concurrency caused both transitions fail but handler is still acquirable. re-schedule this task.
                        housekeepingExecutor.execute(this);
                    }
                    break;
                case LEAK:
                    if (handler.isLeak(configuration.leakTimeout()) && handler.setState(CHECKED_OUT, FLUSH)) {
                        flushHandler(handler);
                    }
                    break;
                case IDLE:
                    if (allConnections.size() > configuration.minSize() && handler.setState(CHECKED_IN, FLUSH)) {
                        flushHandler(handler);
                    }
                    break;
                case INVALID:
                    fireBeforeConnectionValidation(listeners, handler);
                    if (handler.setState(CHECKED_IN, VALIDATION)) {
                        if (handler.isValid() && handler.setState(VALIDATION, CHECKED_IN)) {
                            fireOnConnectionValid(listeners, handler);
                        } else {
                            handler.setState(VALIDATION, FLUSH);
                            fireOnConnectionInvalid(listeners, handler);
                            flushHandler(handler);
                        }
                    }
                    break;
                default:
            }
        }

        private void flushHandler(ConnectionHandler handler) {
            allConnections.remove(handler);
            synchronizer.releaseConditional();
            metricsRepository.afterConnectionFlush();
            fireOnConnectionFlush(listeners, handler);
            housekeepingExecutor.execute(new DestroyConnectionTask(handler));
        }

        private void afterFlush(XbibDataSource.FlushMode mode) {
            switch (mode) {
                case ALL:
                case GRACEFUL:
                case INVALID:
                case LEAK:
                case FILL:
                    // refill to minSize
                    housekeepingExecutor.execute(new FillTask());
                    break;
                case IDLE:
                    break;
                default:
                    fireOnWarning(listeners, "Unsupported Flush mode " + mode);
            }
        }
    }

    // --- fill task //

    private final class FillTask implements Runnable {

        @Override
        public void run() {
            for (int n = configuration.minSize() - allConnections.size(); n > 0; n--) {
                housekeepingExecutor.executeNow(new CreateConnectionTask());
            }
        }
    }

    // --- leak detection //

    private final class LeakTask implements Runnable {

        @Override
        public void run() {
            housekeepingExecutor.schedule(this, configuration.leakTimeout().toNanos(), NANOSECONDS);

            for (ConnectionHandler handler : allConnections) {
                housekeepingExecutor.execute(new LeakConnectionTask(handler));
            }
        }

        private class LeakConnectionTask implements Runnable {

            private final ConnectionHandler handler;

            LeakConnectionTask(ConnectionHandler handler) {
                this.handler = handler;
            }

            @Override
            public void run() {
                fireBeforeConnectionLeak(listeners, handler);
                if (handler.isLeak(configuration.leakTimeout())) {
                    metricsRepository.afterLeakDetection();
                    fireOnConnectionLeak(listeners, handler);
                }
            }
        }
    }

    // --- validation //

    private final class ValidationTask implements Runnable {

        @Override
        public void run() {
            housekeepingExecutor.schedule(this, configuration.validationTimeout().toNanos(), NANOSECONDS);

            for (ConnectionHandler handler : allConnections) {
                housekeepingExecutor.execute(new ValidateConnectionTask(handler));
            }
        }

        private class ValidateConnectionTask implements Runnable {

            private final ConnectionHandler handler;

            ValidateConnectionTask(ConnectionHandler handler) {
                this.handler = handler;
            }

            @Override
            public void run() {
                if (handler.setState(CHECKED_IN, VALIDATION)) {
                    performValidation(handler, CHECKED_IN);
                }
            }
        }
    }

    // --- reap //

    private final class ReapTask implements Runnable {

        @Override
        public void run() {
            housekeepingExecutor.schedule(this, configuration.reapTimeout().toNanos(), NANOSECONDS);

            // reset the thead local cache
            localCache.reset();

            for (ConnectionHandler handler : allConnections) {
                housekeepingExecutor.execute(new ReapConnectionTask(handler));
            }
        }

        private class ReapConnectionTask implements Runnable {

            private final ConnectionHandler handler;

            ReapConnectionTask(ConnectionHandler handler) {
                this.handler = handler;
            }

            @Override
            public void run() {
                fireBeforeConnectionReap(listeners, handler);
                if (allConnections.size() > configuration.minSize() && handler.setState(CHECKED_IN, FLUSH)) {
                    if (handler.isIdle(configuration.reapTimeout())) {
                        removeFromPool(handler);
                        metricsRepository.afterConnectionReap();
                        fireOnConnectionReap(listeners, handler);
                    } else {
                        handler.setState(CHECKED_IN);
                        // for debug, something like: fireOnWarning( listeners,  "Connection " + handler.getConnection() + " used recently. Do not reap!" );
                    }
                }
            }
        }
    }

    // --- destroy //

    private final class DestroyConnectionTask implements Runnable {

        private final ConnectionHandler handler;

        DestroyConnectionTask(ConnectionHandler handler) {
            this.handler = handler;
        }

        @Override
        public void run() {
            fireBeforeConnectionDestroy(listeners, handler);
            try {
                fireOnConnectionDestroyInterceptor(interceptors, handler);
                handler.closeConnection();
            } catch (SQLException e) {
                fireOnWarning(listeners, e);
            }
            metricsRepository.afterConnectionDestroy();
            fireOnConnectionDestroy(listeners, handler);
        }
    }
}
