package org.xbib.jdbc.pool;

import java.io.IOException;
import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.SQLFeatureNotSupportedException;
import java.util.Collection;
import java.util.List;
import java.util.logging.Logger;
import org.xbib.jdbc.pool.api.XbibDataSource;
import org.xbib.jdbc.pool.api.XbibDataSourceListener;
import org.xbib.jdbc.pool.api.XbibDataSourceMetrics;
import org.xbib.jdbc.pool.api.XbibPoolInterceptor;
import org.xbib.jdbc.pool.api.configuration.XbibDataSourceConfiguration;
import static java.util.Collections.emptyList;

public final class DataSource implements XbibDataSource {

    private final XbibDataSourceConfiguration configuration;
    private final Pool connectionPool;

    public DataSource(XbibDataSourceConfiguration dataSourceConfiguration, XbibDataSourceListener... listeners) {
        configuration = dataSourceConfiguration;
        if (configuration.dataSourceImplementation() == XbibDataSourceConfiguration.DataSourceImplementation.XBIB_POOLLESS) {
            connectionPool = new Poolless(dataSourceConfiguration.connectionPoolConfiguration(), listeners);
        } else {
            connectionPool = new ConnectionPool(dataSourceConfiguration.connectionPoolConfiguration(), listeners);
        }

        dataSourceConfiguration.registerMetricsEnabledListener(connectionPool);
        connectionPool.onMetricsEnabled(dataSourceConfiguration.metricsEnabled());
        connectionPool.init();
    }

    @Override
    public List<XbibPoolInterceptor> getPoolInterceptors() {
        return connectionPool.getPoolInterceptors();
    }

    @Override
    public void setPoolInterceptors(Collection<? extends XbibPoolInterceptor> interceptors) {
        connectionPool.setPoolInterceptors(interceptors == null ? emptyList() : interceptors);
    }

    @Override
    public XbibDataSourceConfiguration getConfiguration() {
        return configuration;
    }

    @Override
    public XbibDataSourceMetrics getMetrics() {
        return connectionPool.getMetrics();
    }

    @Override
    public void flush(FlushMode mode) {
        connectionPool.flushPool(mode);
    }

    @Override
    public boolean isHealthy(boolean newConnection) throws SQLException {
        return connectionPool.isHealthy(newConnection);
    }

    @Override
    public void close() throws IOException {
        connectionPool.close();
    }

    @Override
    public Connection getConnection() throws SQLException {
        return connectionPool.getConnection();
    }

    @Override
    public Connection getConnection(String username, String password) throws SQLException {
        throw new SQLException("username and password combination invalid on a pooled data source!");
    }

    @Override
    public <T> T unwrap(Class<T> target) throws SQLException {
        return target.cast(this);
    }

    @Override
    public boolean isWrapperFor(Class<?> target) throws SQLException {
        return target.isInstance(this);
    }

    @Override
    public PrintWriter getLogWriter() throws SQLException {
        return null;
    }

    @Override
    public void setLogWriter(PrintWriter out) throws SQLException {
        // no-op
    }

    @Override
    public int getLoginTimeout() throws SQLException {
        return 0;
    }

    @Override
    public void setLoginTimeout(int seconds) throws SQLException {
        // no-op
    }

    @Override
    public Logger getParentLogger() throws SQLFeatureNotSupportedException {
        throw new SQLFeatureNotSupportedException("Not Supported");
    }
}
