package org.xbib.jdbc.pool;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.LongAccumulator;
import java.util.function.Function;
import javax.sql.XAConnection;
import org.xbib.jdbc.pool.MetricsRepository.EmptyMetricsRepository;
import org.xbib.jdbc.pool.api.XbibDataSource;
import org.xbib.jdbc.pool.api.XbibDataSourceListener;
import org.xbib.jdbc.pool.api.XbibPoolInterceptor;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration;
import org.xbib.jdbc.pool.api.transaction.TransactionIntegration;
import org.xbib.jdbc.pool.util.StampedCopyOnWriteArrayList;
import org.xbib.jdbc.pool.util.XbibSynchronizer;
import static java.lang.Integer.toHexString;
import static java.lang.System.identityHashCode;
import static java.lang.System.nanoTime;
import static java.lang.Thread.currentThread;
import static java.util.Arrays.copyOfRange;
import static java.util.Collections.unmodifiableList;
import static java.util.stream.Collectors.joining;
import static java.util.stream.Collectors.toList;
import static org.xbib.jdbc.pool.ConnectionHandler.State.CHECKED_OUT;
import static org.xbib.jdbc.pool.ConnectionHandler.State.FLUSH;
import static org.xbib.jdbc.pool.ConnectionHandler.State.VALIDATION;
import static org.xbib.jdbc.pool.api.XbibDataSource.FlushMode.ALL;
import static org.xbib.jdbc.pool.api.XbibDataSource.FlushMode.LEAK;
import static org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.MultipleAcquisitionAction.OFF;
import static org.xbib.jdbc.pool.util.InterceptorHelper.fireOnConnectionAcquiredInterceptor;
import static org.xbib.jdbc.pool.util.InterceptorHelper.fireOnConnectionCreateInterceptor;
import static org.xbib.jdbc.pool.util.InterceptorHelper.fireOnConnectionDestroyInterceptor;
import static org.xbib.jdbc.pool.util.InterceptorHelper.fireOnConnectionReturnInterceptor;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionAcquire;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionCreation;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionDestroy;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionFlush;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionReturn;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireBeforeConnectionValidation;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionAcquired;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionCreation;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionDestroy;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionFlush;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionInvalid;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionPooled;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnConnectionValid;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnInfo;
import static org.xbib.jdbc.pool.util.ListenerHelper.fireOnWarning;

/**
 * Alternative implementation of ConnectionPool for the special case of flush-on-close (and min-size == 0)
 * In particular, this removes the need for the executor.
 * Also, there is no thread-local connection cache as connections are not reused
 */
public final class Poolless implements Pool {

    private final XbibConnectionPoolConfiguration configuration;
    private final XbibDataSourceListener[] listeners;

    private final StampedCopyOnWriteArrayList<ConnectionHandler> allConnections;

    private final XbibSynchronizer synchronizer;
    private final ConnectionFactory connectionFactory;
    private final TransactionIntegration transactionIntegration;

    private final LongAccumulator maxUsed = new LongAccumulator(Math::max, Long.MIN_VALUE);
    private final AtomicInteger activeCount = new AtomicInteger();

    private List<XbibPoolInterceptor> interceptors;
    private MetricsRepository metricsRepository;
    private volatile boolean shutdown;

    public Poolless(XbibConnectionPoolConfiguration configuration, XbibDataSourceListener... listeners) {
        this.configuration = configuration;
        this.listeners = listeners;

        allConnections = new StampedCopyOnWriteArrayList<>(ConnectionHandler.class);

        synchronizer = new XbibSynchronizer();
        connectionFactory = new ConnectionFactory(configuration.connectionFactoryConfiguration(), listeners);
        transactionIntegration = configuration.transactionIntegration();
    }

    public void init() {
        if (!configuration.maxLifetime().isZero()) {
            fireOnInfo(listeners, "Max lifetime not supported in pool-less mode");
        }
        if (configuration.validateOnBorrow()) {
            fireOnInfo(listeners, "Validation on borrow is not supported in pool-less mode");
        }
        if (!configuration.idleValidationTimeout().isZero()) {
            fireOnInfo(listeners, "Idle validation not supported in pool-less mode");
        }
        if (!configuration.leakTimeout().isZero()) {
            fireOnInfo(listeners, "Leak detection not pro-active in pool-less mode");
        }
        if (!configuration.reapTimeout().isZero()) {
            fireOnInfo(listeners, "Connection reap not supported in pool-less mode");
        }
        if (configuration.initialSize() != 0) {
            fireOnInfo(listeners, "Initial size is zero in pool-less mode");
        }
        if (configuration.minSize() != 0) {
            fireOnInfo(listeners, "Min size always zero in pool-less mode");
        }
        transactionIntegration.addResourceRecoveryFactory(connectionFactory.hasRecoveryCredentials() ? connectionFactory : this);
    }

    public XbibConnectionPoolConfiguration getConfiguration() {
        return configuration;
    }

    public int defaultJdbcIsolationLevel() {
        return connectionFactory.defaultJdbcIsolationLevel();
    }

    public XbibDataSourceListener[] getListeners() {
        return listeners;
    }

    public List<XbibPoolInterceptor> getPoolInterceptors() {
        return unmodifiableList(interceptors);
    }

    @SuppressWarnings({"StringConcatenation", "SingleCharacterStringConcatenation"})
    public void setPoolInterceptors(Collection<? extends XbibPoolInterceptor> list) {
        if (list.stream().anyMatch(i -> i.getPriority() < 0)) {
            throw new IllegalArgumentException("Negative priority values on XbibPoolInterceptor are reserved.");
        }
        if (list.isEmpty() && (interceptors == null || interceptors.isEmpty())) {
            return;
        }
        interceptors = list.stream().sorted(XbibPoolInterceptor.DEFAULT_COMPARATOR).collect(toList());

        Function<XbibPoolInterceptor, String> interceptorName = i -> i.getClass().getName() + "@" + toHexString(identityHashCode(i)) + " (priority " + i.getPriority() + ")";
        fireOnInfo(listeners, "Pool interceptors: " + interceptors.stream().map(interceptorName).collect(joining(" >>> ", "[", "]")));
    }

    // --- //

    @Override
    public boolean isRecoverable() {
        return connectionFactory.isRecoverable();
    }

    @Override
    public XAConnection getRecoveryConnection() throws SQLException {
        long stamp = beforeAcquire();
        checkMultipleAcquisition();

        ConnectionHandler checkedOutHandler = handlerFromSharedCache();
        fireOnConnectionAcquiredInterceptor(interceptors, checkedOutHandler);
        afterAcquire(stamp, checkedOutHandler, false);
        return checkedOutHandler.xaConnectionWrapper();
    }

    @Override
    public void close() {
        transactionIntegration.removeResourceRecoveryFactory(connectionFactory.hasRecoveryCredentials() ? connectionFactory : this);
        shutdown = true;

        for (ConnectionHandler handler : allConnections) {
            handler.setState(FLUSH);
            destroyConnection(handler);
        }
        allConnections.clear();

        synchronizer.release(synchronizer.getQueueLength());
    }

    private long beforeAcquire() throws SQLException {
        fireBeforeConnectionAcquire(listeners);
        if (shutdown) {
            throw new SQLException("This pool is closed and does not handle any more connections!");
        }
        return metricsRepository.beforeConnectionAcquire();
    }

    private void checkMultipleAcquisition() throws SQLException {
        if (configuration.multipleAcquisition() != OFF) {
            for (ConnectionHandler handler : allConnections) {
                if (handler.getHoldingThread() == currentThread()) {
                    switch (configuration.multipleAcquisition()) {
                        case STRICT:
                            throw new SQLException("Acquisition of multiple connections by the same Thread.");
                        case WARN:
                            fireOnWarning(listeners, "Acquisition of multiple connections by the same Thread. This can lead to pool exhaustion and eventually a deadlock!");
                    }
                    break;
                }
            }
        }
    }

    public Connection getConnection() throws SQLException {
        long stamp = beforeAcquire();

        ConnectionHandler checkedOutHandler = handlerFromTransaction();
        if (checkedOutHandler != null) {
            // AG-140 - If associate throws here is fine, it's assumed the synchronization that returns the connection has been registered
            transactionIntegration.associate(checkedOutHandler, checkedOutHandler.getXaResource());
            afterAcquire(stamp, checkedOutHandler, false);
            return checkedOutHandler.connectionWrapper();
        }
        checkMultipleAcquisition();

        try {
            checkedOutHandler = handlerFromSharedCache();
            transactionIntegration.associate(checkedOutHandler, checkedOutHandler.getXaResource());
            fireOnConnectionAcquiredInterceptor(interceptors, checkedOutHandler);
            afterAcquire(stamp, checkedOutHandler, true);
            return checkedOutHandler.connectionWrapper();
        } catch (Throwable t) {
            if (checkedOutHandler != null) {
                // AG-140 - Flush handler to prevent leak
                flushHandler(checkedOutHandler);
            }
            throw t;
        }
    }

    private ConnectionHandler handlerFromTransaction() throws SQLException {
        return (ConnectionHandler) transactionIntegration.getTransactionAware();
    }

    private ConnectionHandler handlerFromSharedCache() throws SQLException {
        long remaining = configuration.acquisitionTimeout().toNanos();
        remaining = remaining > 0 ? remaining : Long.MAX_VALUE;
        try {
            for (; ; ) {
                // Try to get a "token" to create a new connection
                if (activeCount.incrementAndGet() <= configuration.maxSize()) {
                    try {
                        return createConnection();
                    } catch (SQLException e) {
                        activeCount.decrementAndGet();
                        throw e;
                    }
                } else {
                    activeCount.decrementAndGet();
                }

                // Pool full, will have to wait for a connection to be returned
                long synchronizationStamp = synchronizer.getStamp();
                long start = nanoTime();
                if (remaining < 0 || !synchronizer.tryAcquireNanos(synchronizationStamp, remaining)) {
                    throw new SQLException("Sorry, acquisition timeout!");
                }
                if (shutdown) {
                    throw new SQLException("Can't create new connection as the pool is shutting down");
                }
                remaining -= nanoTime() - start;
            }
        } catch (InterruptedException e) {
            currentThread().interrupt();
            throw new SQLException("Interrupted while acquiring");
        }
    }

    @SuppressWarnings("SingleCharacterStringConcatenation")
    private void afterAcquire(long metricsStamp, ConnectionHandler checkedOutHandler, boolean verifyEnlistment) throws SQLException {
        metricsRepository.afterConnectionAcquire(metricsStamp);
        fireOnConnectionAcquired(listeners, checkedOutHandler);

        if (verifyEnlistment && !checkedOutHandler.isEnlisted()) {
            switch (configuration.transactionRequirement()) {
                case STRICT:
                    returnConnectionHandler(checkedOutHandler);
                    throw new SQLException("Connection acquired without transaction.");
                case WARN:
                    fireOnWarning(listeners, new SQLException("Connection acquired without transaction."));
            }
        }
        if (!configuration.leakTimeout().isZero() || configuration.multipleAcquisition() != OFF) {
            if (checkedOutHandler.getHoldingThread() != null && checkedOutHandler.getHoldingThread() != currentThread()) {
                Throwable warn = new Throwable("Shared connection between threads '" + checkedOutHandler.getHoldingThread().getName() + "' and '" + currentThread().getName() + "'");
                warn.setStackTrace(checkedOutHandler.getHoldingThread().getStackTrace());
                fireOnWarning(listeners, warn);
            }
            checkedOutHandler.setHoldingThread(currentThread());
            checkedOutHandler.touch();
            if (configuration.enhancedLeakReport()) {
                StackTraceElement[] stackTrace = currentThread().getStackTrace();
                checkedOutHandler.setAcquisitionStackTrace(copyOfRange(stackTrace, 4, stackTrace.length));
            }
        }
    }

    public void returnConnectionHandler(ConnectionHandler handler) throws SQLException {
        fireBeforeConnectionReturn(listeners, handler);
        try {
            if (!transactionIntegration.disassociate(handler)) {
                return;
            }
        } catch (Throwable ignored) {
        }

        fireOnConnectionReturnInterceptor(interceptors, handler);
        flushHandler(handler);
    }

    @Override
    public void onMetricsEnabled(boolean metricsEnabled) {
        metricsRepository = metricsEnabled ? new DefaultMetricsRepository(this) : new EmptyMetricsRepository();
    }

    public MetricsRepository getMetrics() {
        return metricsRepository;
    }

    public long activeCount() {
        return activeCount.get();
    }

    public long availableCount() {
        return configuration.maxSize() - activeCount.get();
    }

    public long maxUsedCount() {
        return maxUsed.get();
    }

    public void resetMaxUsedCount() {
        maxUsed.reset();
    }

    public long awaitingCount() {
        return synchronizer.getQueueLength();
    }

    @Override
    public boolean isHealthy(boolean newConnection) throws SQLException {
        // the main difference here is that connect will not block (and potentially timeout) on full pool
        if (newConnection) {
            activeCount.incrementAndGet();
        }
        ConnectionHandler healthHandler = newConnection ? createConnection() : handlerFromSharedCache();

        try {
            fireBeforeConnectionValidation(listeners, healthHandler);
            if (healthHandler.setState(CHECKED_OUT, VALIDATION) && healthHandler.isValid() && healthHandler.setState(VALIDATION, CHECKED_OUT)) {
                fireOnConnectionValid(listeners, healthHandler);
                return true;
            } else {
                metricsRepository.afterConnectionInvalid();
                fireOnConnectionInvalid(listeners, healthHandler);
                return false;
            }
        } finally {
            returnConnectionHandler(healthHandler);
        }
    }

    private ConnectionHandler createConnection() throws SQLException {
        fireBeforeConnectionCreation(listeners);
        long metricsStamp = metricsRepository.beforeConnectionCreation();

        try {
            ConnectionHandler handler = new ConnectionHandler(connectionFactory.createConnection(), this);
            metricsRepository.afterConnectionCreation(metricsStamp);

            fireOnConnectionCreation(listeners, handler);
            fireOnConnectionCreateInterceptor(interceptors, handler);

            handler.setState(CHECKED_OUT);
            allConnections.add(handler);

            maxUsed.accumulate(allConnections.size());
            fireOnConnectionPooled(listeners, handler);

            return handler;
        } catch (SQLException e) {
            fireOnWarning(listeners, e);
            throw e;
        }
    }

    public void flushPool(XbibDataSource.FlushMode mode) {
        if (mode == ALL) {
            for (ConnectionHandler handler : allConnections) {
                fireBeforeConnectionFlush(listeners, handler);
                flushHandler(handler);
            }
        } else if (mode == LEAK) {
            for (ConnectionHandler handler : allConnections) {
                if (handler.isLeak(configuration.leakTimeout())) {
                    fireBeforeConnectionFlush(listeners, handler);
                    flushHandler(handler);
                }
            }
        }
    }

    private void flushHandler(ConnectionHandler handler) {
        handler.setState(FLUSH);
        allConnections.remove(handler);
        activeCount.decrementAndGet();
        synchronizer.releaseConditional();
        metricsRepository.afterConnectionFlush();
        fireOnConnectionFlush(listeners, handler);
        destroyConnection(handler);
    }

    private void destroyConnection(ConnectionHandler handler) {
        fireBeforeConnectionDestroy(listeners, handler);
        try {
            fireOnConnectionDestroyInterceptor(interceptors, handler);
            handler.closeConnection();
        } catch (SQLException e) {
            fireOnWarning(listeners, e);
        }
        metricsRepository.afterConnectionDestroy();
        fireOnConnectionDestroy(listeners, handler);
    }
}
