package org.xbib.jdbc.pool.api;

import java.io.Closeable;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.function.Supplier;
import javax.sql.DataSource;
import org.xbib.jdbc.pool.api.configuration.XbibDataSourceConfiguration;
import static java.util.ServiceLoader.load;

/**
 * Extension of the DataSource interface that exposes some of its internals.
 */
public interface XbibDataSource extends DataSource, Closeable {

    /**
     * Create an XbibDataSource from a supplier of the configuration.
     */
    static XbibDataSource from(Supplier<XbibDataSourceConfiguration> configurationSupplier, XbibDataSourceListener... listeners) throws SQLException {
        return from(configurationSupplier.get(), listeners);
    }

    /**
     * Create an XbibDataSource from configuration.
     */
    static XbibDataSource from(XbibDataSourceConfiguration configuration, XbibDataSourceListener... listeners) throws SQLException {
        for (XbibDataSourceProvider provider : load(XbibDataSourceProvider.class, XbibDataSourceProvider.class.getClassLoader())) {
            XbibDataSource implementation = provider.getDataSource(configuration, listeners);
            if (implementation != null) {
                return implementation;
            }
        }

        // Fall back to load the implementation using reflection
        try {
            Class<? extends XbibDataSource> dataSourceClass = XbibDataSource.class.getClassLoader().loadClass(configuration.dataSourceImplementation().className()).asSubclass(XbibDataSource.class);
            Constructor<? extends XbibDataSource> dataSourceConstructor = dataSourceClass.getConstructor(XbibDataSourceConfiguration.class, XbibDataSourceListener[].class);
            return dataSourceConstructor.newInstance(configuration, listeners);
        } catch (ClassNotFoundException | NoSuchMethodException | IllegalAccessException | InvocationTargetException |
                 InstantiationException e) {
            throw new SQLException("Could not load the required implementation", e);
        }
    }

    /**
     * Allows inspection of the configuration. Some properties allow read / write.
     */
    XbibDataSourceConfiguration getConfiguration();

    /**
     * Allows access to metrics. If metrics are not enabled, returns default values.
     */
    XbibDataSourceMetrics getMetrics();

    /**
     * Performs a flush action on the connections of the pool.
     */
    void flush(FlushMode mode);

    /**
     * Get the list of pool interceptors. Interceptors are sorted from high to low priority.
     */
    List<XbibPoolInterceptor> getPoolInterceptors();

    /**
     * Sets pool interceptors.
     */
    void setPoolInterceptors(Collection<? extends XbibPoolInterceptor> interceptors);

    /**
     * Performs a health check. The newConnection parameter determines that a new database connection is established for this purpose, otherwise attempts to get a connection from the pool.
     * <p>
     * WARNING: Using a new connection may cause the size of the pool to go over max-size.
     */
    default boolean isHealthy(boolean newConnection) throws SQLException {
        return true;
    }

    /**
     * Modes supported on the flush operation.
     */
    enum FlushMode {

        /**
         * All connections are flushed right away.
         */
        ALL,

        /**
         * Idle connections are flushed.
         */
        IDLE,

        /**
         * Performs on-demand validation of idle connections.
         */
        INVALID,

        /**
         * Active connections are flushed on return. Idle connections are flushed immediately.
         */
        GRACEFUL,

        /**
         * Flushes connections that have been in use for longer than the specified leak timeout.
         */
        LEAK,

        /**
         * Creates connections to met the minimum size of the pool.
         * Used after and increase of minimum size, to make that change effective immediately.
         */
        FILL
    }
}
