package org.xbib.jdbc.pool.api;

import java.sql.Connection;
import java.util.Comparator;

/**
 * Callback interface for pool actions.
 * These differ from the Listener in a few different ways:
 * They do not have access to the raw Connection
 * The invoke order is dependent on the operation (incoming / outgoing)
 * Consistent with the transaction.
 */
public interface XbibPoolInterceptor {

    /**
     * Uses interceptor priority, followed by {@link Class#getName()} to ensure a consistent ordering.
     */
    Comparator<XbibPoolInterceptor> DEFAULT_COMPARATOR =
            Comparator.comparingInt(XbibPoolInterceptor::getPriority).thenComparing(i -> i.getClass().getName());

    /**
     * This callback is invoked when a new connection is created.
     */
    default void onConnectionCreate(Connection connection) {
    }

    /**
     * This callback is invoked when a connection is successfully acquired.
     * When in a transactional environment this is invoked only once for multiple acquire calls within the same transaction, before the connection is associated.
     */
    default void onConnectionAcquire(Connection connection) {
    }

    /**
     * This callback is invoked before a connection is returned to the pool.
     * When in a transactional environment this is invoked only once for each transaction, after commit / rollback.
     * This callback runs after reset, allowing connections to be in the pool in a different state than the described on the configuration.
     */
    default void onConnectionReturn(Connection connection) {
    }

    /**
     * This callback is invoked just before a connection is destroyed.
     */
    default void onConnectionDestroy(Connection connection) {
    }

    /**
     * Allows a ordering between multiple interceptors.
     * Lower priority are executed first. Negative values are reserved.
     */
    default int getPriority() {
        return 0;
    }
}
