package org.xbib.jdbc.pool.api.configuration.supplier;

import java.time.Duration;
import java.util.function.Function;
import java.util.function.Supplier;
import org.xbib.jdbc.pool.api.cache.ConnectionCache;
import org.xbib.jdbc.pool.api.cache.LocalConnectionCache;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionFactoryConfiguration;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.MultipleAcquisitionAction;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.TransactionRequirement;
import org.xbib.jdbc.pool.api.transaction.TransactionIntegration;
import static java.lang.Integer.MAX_VALUE;
import static java.time.Duration.ZERO;
import static org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.ConnectionValidator.emptyValidator;
import static org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.ExceptionSorter.emptyExceptionSorter;
import static org.xbib.jdbc.pool.api.transaction.TransactionIntegration.none;

/**
 * Builder of XbibConnectionPoolConfiguration.
 */
@SuppressWarnings({"PackageVisibleField", "WeakerAccess"})
public class XbibConnectionPoolConfigurationSupplier implements Supplier<XbibConnectionPoolConfiguration> {

    XbibConnectionFactoryConfiguration connectionFactoryConfiguration;

    ConnectionCache connectionCache = LocalConnectionCache.single();
    TransactionIntegration transactionIntegration = none();
    TransactionRequirement transactionRequirement = TransactionRequirement.OFF;
    MultipleAcquisitionAction multipleAcquisitionAction = MultipleAcquisitionAction.OFF;
    boolean enhancedLeakReport;
    boolean flushOnClose;
    int initialSize;
    volatile int minSize;
    volatile int maxSize = MAX_VALUE;
    boolean validateOnBorrow;
    XbibConnectionPoolConfiguration.ConnectionValidator connectionValidator = emptyValidator();
    XbibConnectionPoolConfiguration.ExceptionSorter exceptionSorter = emptyExceptionSorter();
    Duration idleValidationTimeout = ZERO;
    Duration leakTimeout = ZERO;
    Duration validationTimeout = ZERO;
    Duration reapTimeout = ZERO;
    Duration maxLifetime = ZERO;
    volatile Duration acquisitionTimeout = ZERO;

    private volatile boolean lock;

    private XbibConnectionFactoryConfigurationSupplier connectionFactoryConfigurationSupplier = new XbibConnectionFactoryConfigurationSupplier();

    public XbibConnectionPoolConfigurationSupplier() {
        lock = false;
    }

    public XbibConnectionPoolConfigurationSupplier(XbibConnectionPoolConfiguration existingConfiguration) {
        this();
        if (existingConfiguration == null) {
            return;
        }
        connectionFactoryConfigurationSupplier = new XbibConnectionFactoryConfigurationSupplier(existingConfiguration.connectionFactoryConfiguration());
        connectionCache = existingConfiguration.connectionCache();
        transactionIntegration = existingConfiguration.transactionIntegration();
        transactionRequirement = existingConfiguration.transactionRequirement();
        multipleAcquisitionAction = existingConfiguration.multipleAcquisition();
        flushOnClose = existingConfiguration.flushOnClose();
        enhancedLeakReport = existingConfiguration.enhancedLeakReport();
        initialSize = existingConfiguration.initialSize();
        minSize = existingConfiguration.minSize();
        maxSize = existingConfiguration.maxSize();
        validateOnBorrow = existingConfiguration.validateOnBorrow();
        connectionValidator = existingConfiguration.connectionValidator();
        exceptionSorter = existingConfiguration.exceptionSorter();
        idleValidationTimeout = existingConfiguration.idleValidationTimeout();
        leakTimeout = existingConfiguration.leakTimeout();
        validationTimeout = existingConfiguration.validationTimeout();
        reapTimeout = existingConfiguration.reapTimeout();
        maxLifetime = existingConfiguration.maxLifetime();
        acquisitionTimeout = existingConfiguration.acquisitionTimeout();
    }

    private void checkLock() {
        if (lock) {
            throw new IllegalStateException("Attempt to modify an immutable configuration");
        }
    }

    /**
     * Sets the configuration for the connection factory.
     */
    private XbibConnectionPoolConfigurationSupplier connectionFactoryConfiguration(XbibConnectionFactoryConfiguration configuration) {
        checkLock();
        connectionFactoryConfigurationSupplier = new XbibConnectionFactoryConfigurationSupplier(configuration);
        return this;
    }

    /**
     * Sets the configuration for the connection factory.
     */
    public XbibConnectionPoolConfigurationSupplier connectionFactoryConfiguration(Supplier<? extends XbibConnectionFactoryConfiguration> supplier) {
        return connectionFactoryConfiguration(supplier.get());
    }

    /**
     * Modifies the configuration of the connection pool.
     */
    public XbibConnectionPoolConfigurationSupplier connectionFactoryConfiguration(Function<? super XbibConnectionFactoryConfigurationSupplier, ? extends XbibConnectionFactoryConfigurationSupplier> function) {
        return connectionFactoryConfiguration(function.apply(connectionFactoryConfigurationSupplier));
    }

    /**
     * Allows access to the configuration builder for the connection pool.
     */
    public XbibConnectionFactoryConfigurationSupplier connectionFactoryConfiguration() {
        return connectionFactoryConfigurationSupplier;
    }

    // --- //

    /**
     * Sets the connection cache implementation. Default is {@link LocalConnectionCache#single}.
     */
    public XbibConnectionPoolConfigurationSupplier connectionCache(ConnectionCache cache) {
        checkLock();
        connectionCache = cache;
        return this;
    }

    /**
     * Sets the transaction integration instance to use. Default is {@link TransactionIntegration#none()}.
     */
    public XbibConnectionPoolConfigurationSupplier transactionIntegration(TransactionIntegration integration) {
        checkLock();
        transactionIntegration = integration;
        return this;
    }

    /**
     * Sets the transaction requirements for the pool. Default is {@link TransactionRequirement#OFF}.
     */
    public XbibConnectionPoolConfigurationSupplier transactionRequirement(TransactionRequirement requirement) {
        checkLock();
        transactionRequirement = requirement;
        return this;
    }

    /**
     * Enables enhanced leak report.
     */
    public XbibConnectionPoolConfigurationSupplier enhancedLeakReport() {
        return enhancedLeakReport(true);
    }

    /**
     * Sets the behaviour of the pool when a thread tries to acquire multiple connections. Default is {@link MultipleAcquisitionAction#OFF}
     */
    public XbibConnectionPoolConfigurationSupplier multipleAcquisition(MultipleAcquisitionAction action) {
        checkLock();
        multipleAcquisitionAction = action;
        return this;
    }

    /**
     * Enables or disables enhanced leak report. Default is false.
     */
    public XbibConnectionPoolConfigurationSupplier enhancedLeakReport(boolean enhanced) {
        checkLock();
        enhancedLeakReport = enhanced;
        return this;
    }

    /**
     * Enables flushing of connections on close.
     */
    public XbibConnectionPoolConfigurationSupplier flushOnClose() {
        return flushOnClose(true);
    }

    /**
     * Enables or disables flushing of connections on close. Default is false.
     */
    public XbibConnectionPoolConfigurationSupplier flushOnClose(boolean flush) {
        checkLock();
        flushOnClose = flush;
        return this;
    }

    /**
     * Sets the number of connections when the pool starts. Must not be negative. Default is zero.
     */
    public XbibConnectionPoolConfigurationSupplier initialSize(int size) {
        checkLock();
        initialSize = size;
        return this;
    }

    /**
     * Sets the minimum number of connections on the pool. Must not be negative and smaller than max. Default is zero.
     */
    public XbibConnectionPoolConfigurationSupplier minSize(int size) {
        checkLock();
        minSize = size;
        return this;
    }

    /**
     * Sets the maximum number of connections on the pool. Must not be negative. Required.
     */
    public XbibConnectionPoolConfigurationSupplier maxSize(int size) {
        checkLock();
        maxSize = size;
        return this;
    }

    /**
     * Enables validation on borrow. Default is false.
     */
    public XbibConnectionPoolConfigurationSupplier validateOnBorrow(boolean validate) {
        checkLock();
        validateOnBorrow = validate;
        return this;
    }

    /**
     * Sets the connection validation method. Default {@link XbibConnectionPoolConfiguration.ConnectionValidator#emptyValidator()}
     */
    public XbibConnectionPoolConfigurationSupplier connectionValidator(XbibConnectionPoolConfiguration.ConnectionValidator validator) {
        checkLock();
        connectionValidator = validator;
        return this;
    }

    /**
     * Sets the exception sorter. Default {@link XbibConnectionPoolConfiguration.ExceptionSorter#emptyExceptionSorter()}
     */
    public XbibConnectionPoolConfigurationSupplier exceptionSorter(XbibConnectionPoolConfiguration.ExceptionSorter sorter) {
        checkLock();
        exceptionSorter = sorter;
        return this;
    }

    /**
     * Sets the duration of the acquisition timeout. Default is {@link Duration#ZERO} meaning that a thread will wait indefinitely.
     */
    public XbibConnectionPoolConfigurationSupplier acquisitionTimeout(Duration timeout) {
        checkLock();
        acquisitionTimeout = timeout;
        return this;
    }

    /**
     * Sets the duration of idle time for foreground validation to be executed. Default is {@link Duration#ZERO} meaning that this feature is disabled.
     */
    public XbibConnectionPoolConfigurationSupplier idleValidationTimeout(Duration timeout) {
        checkLock();
        idleValidationTimeout = timeout;
        return this;
    }

    /**
     * Sets the duration of the leak timeout detection. Default is {@link Duration#ZERO} meaning that this feature is disabled.
     */
    public XbibConnectionPoolConfigurationSupplier leakTimeout(Duration timeout) {
        checkLock();
        leakTimeout = timeout;
        return this;
    }

    /**
     * Sets the duration of background validation interval. Default is {@link Duration#ZERO} meaning that this feature is disabled.
     */
    public XbibConnectionPoolConfigurationSupplier validationTimeout(Duration timeout) {
        checkLock();
        validationTimeout = timeout;
        return this;
    }

    /**
     * Sets the duration for eviction of idle connections. Default is {@link Duration#ZERO} meaning that this feature is disabled.
     */
    public XbibConnectionPoolConfigurationSupplier reapTimeout(Duration timeout) {
        checkLock();
        reapTimeout = timeout;
        return this;
    }

    /**
     * Sets the duration for the lifetime of connections. Default is {@link Duration#ZERO} meaning that this feature is disabled.
     */
    public XbibConnectionPoolConfigurationSupplier maxLifetime(Duration time) {
        checkLock();
        maxLifetime = time;
        return this;
    }

    // --- //

    private void validate() {
        if (lock) {
            throw new IllegalStateException("Configuration already supplied");
        }
        if (maxSize == MAX_VALUE) {
            throw new IllegalArgumentException("max size attribute is mandatory");
        }
        if (maxSize <= 0) {
            throw new IllegalArgumentException("A Positive max size is required");
        }
        if (minSize < 0) {
            throw new IllegalArgumentException("Invalid min size: smaller than 0");
        }
        if (minSize > maxSize) {
            throw new IllegalArgumentException("Invalid min size: greater than max size");
        }
        if (initialSize < 0) {
            throw new IllegalArgumentException("Invalid value for initial size. Must not be negative, and ideally between min size and max size");
        }
        if (acquisitionTimeout.isNegative()) {
            throw new IllegalArgumentException("Acquisition timeout must not be negative");
        }
        if (idleValidationTimeout.isNegative()) {
            throw new IllegalArgumentException("Idle validation timeout must not be negative");
        }
        if (leakTimeout.isNegative()) {
            throw new IllegalArgumentException("Leak detection timeout must not be negative");
        }
        if (reapTimeout.isNegative()) {
            throw new IllegalArgumentException("Reap timeout must not be negative");
        }
        if (maxLifetime.isNegative()) {
            throw new IllegalArgumentException("Max Lifetime must not be negative");
        }
        if (validationTimeout.isNegative()) {
            throw new IllegalArgumentException("Validation timeout must not be negative");
        }
        if (connectionFactoryConfigurationSupplier == null) {
            throw new IllegalArgumentException("Connection factory configuration not defined");
        }
        connectionFactoryConfiguration = connectionFactoryConfigurationSupplier.get();
    }

    @Override
    @SuppressWarnings("ReturnOfInnerClass")
    public XbibConnectionPoolConfiguration get() {
        validate();
        lock = true;

        return new XbibConnectionPoolConfiguration() {

            @Override
            public XbibConnectionFactoryConfiguration connectionFactoryConfiguration() {
                return connectionFactoryConfiguration;
            }

            @Override
            public ConnectionCache connectionCache() {
                return connectionCache;
            }

            @Override
            public TransactionIntegration transactionIntegration() {
                return transactionIntegration;
            }

            @Override
            public TransactionRequirement transactionRequirement() {
                return transactionRequirement;
            }

            @Override
            public boolean enhancedLeakReport() {
                return enhancedLeakReport;
            }

            @Override
            public boolean flushOnClose() {
                return flushOnClose;
            }

            @Override
            public MultipleAcquisitionAction multipleAcquisition() {
                return multipleAcquisitionAction;
            }

            @Override
            public int initialSize() {
                return initialSize;
            }

            @Override
            public int minSize() {
                return minSize;
            }

            @Override
            public void setMinSize(int size) {
                if (size < 0) {
                    throw new IllegalArgumentException("Invalid min size: smaller than 0");
                }
                if (size > maxSize) {
                    throw new IllegalArgumentException("Invalid min size: greater than max size");
                }
                minSize = size;
            }

            @Override
            public int maxSize() {
                return maxSize;
            }

            @Override
            public void setMaxSize(int size) {
                if (size <= 0) {
                    throw new IllegalArgumentException("A Positive max size is required");
                }
                if (size < minSize) {
                    throw new IllegalArgumentException("Invalid max size: smaller than min size");
                }
                maxSize = size;
            }

            @Override
            public Duration acquisitionTimeout() {
                return acquisitionTimeout;
            }

            @Override
            public void setAcquisitionTimeout(Duration timeout) {
                if (timeout.isNegative()) {
                    throw new IllegalArgumentException("Acquisition timeout must not be negative");
                }
                acquisitionTimeout = timeout;
            }

            @Override
            public boolean validateOnBorrow() {
                return validateOnBorrow;
            }

            @Override
            public ConnectionValidator connectionValidator() {
                return connectionValidator;
            }

            @Override
            public ExceptionSorter exceptionSorter() {
                return exceptionSorter;
            }

            @Override
            public Duration idleValidationTimeout() {
                return idleValidationTimeout;
            }

            @Override
            public Duration leakTimeout() {
                return leakTimeout;
            }

            @Override
            public Duration validationTimeout() {
                return validationTimeout;
            }

            @Override
            public Duration reapTimeout() {
                return reapTimeout;
            }

            @Override
            public Duration maxLifetime() {
                return maxLifetime;
            }
        };
    }
}
