package org.xbib.jdbc.pool.api.configuration.supplier;

import java.util.function.Function;
import java.util.function.Supplier;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration;
import org.xbib.jdbc.pool.api.configuration.XbibDataSourceConfiguration;
import org.xbib.jdbc.pool.api.configuration.XbibDataSourceConfiguration.DataSourceImplementation;
import org.xbib.jdbc.pool.api.configuration.XbibDataSourceConfiguration.MetricsEnabledListener;

/**
 * Builder of XbibDataSourceConfiguration.
 */
public class XbibDataSourceConfigurationSupplier implements Supplier<XbibDataSourceConfiguration> {

    XbibConnectionPoolConfiguration connectionPoolConfiguration;

    DataSourceImplementation dataSourceImplementation = DataSourceImplementation.XBIB;
    volatile boolean metrics;
    volatile MetricsEnabledListener metricsEnabledListener;

    private volatile boolean lock;

    private XbibConnectionPoolConfigurationSupplier connectionPoolConfigurationSupplier = new XbibConnectionPoolConfigurationSupplier();

    public XbibDataSourceConfigurationSupplier() {
        lock = false;
    }

    private void checkLock() {
        if (lock) {
            throw new IllegalStateException("Attempt to modify an immutable configuration");
        }
    }

    /**
     * Sets the configuration of the connection pool.
     */
    public XbibDataSourceConfigurationSupplier connectionPoolConfiguration(XbibConnectionPoolConfiguration configuration) {
        checkLock();
        connectionPoolConfigurationSupplier = new XbibConnectionPoolConfigurationSupplier(configuration);
        return this;
    }

    /**
     * Sets the configuration of the connection pool.
     */
    public XbibDataSourceConfigurationSupplier connectionPoolConfiguration(Supplier<? extends XbibConnectionPoolConfiguration> supplier) {
        return connectionPoolConfiguration(supplier.get());
    }

    /**
     * Modifies the configuration of the connection pool.
     */
    public XbibDataSourceConfigurationSupplier connectionPoolConfiguration(Function<? super XbibConnectionPoolConfigurationSupplier, ? extends XbibConnectionPoolConfigurationSupplier> function) {
        return connectionPoolConfiguration(function.apply(connectionPoolConfigurationSupplier));
    }

    /**
     * Allows access to the configuration builder for the connection pool.
     */
    public XbibConnectionPoolConfigurationSupplier connectionPoolConfiguration() {
        return connectionPoolConfigurationSupplier;
    }

    /**
     * Selects the XbibDataSource implementation.
     */
    public XbibDataSourceConfigurationSupplier dataSourceImplementation(DataSourceImplementation implementation) {
        checkLock();
        dataSourceImplementation = implementation;
        return this;
    }

    /**
     * Enables the collection of metrics.
     */
    public XbibDataSourceConfigurationSupplier metricsEnabled() {
        return metricsEnabled(true);
    }

    /**
     * Enables or disables the collection of metrics. The default is false.
     */
    public XbibDataSourceConfigurationSupplier metricsEnabled(boolean metricsEnabled) {
        checkLock();
        metrics = metricsEnabled;
        return this;
    }

    // --- //

    private void validate() {
        if (lock) {
            throw new IllegalStateException("Configuration already supplied");
        }
        if (connectionPoolConfigurationSupplier == null) {
            throw new IllegalArgumentException("Connection pool configuration not defined");
        }
        connectionPoolConfiguration = connectionPoolConfigurationSupplier.get();
    }

    @Override
    @SuppressWarnings("ReturnOfInnerClass")
    public XbibDataSourceConfiguration get() {
        validate();
        lock = true;

        return new XbibDataSourceConfiguration() {

            @Override
            public XbibConnectionPoolConfiguration connectionPoolConfiguration() {
                return connectionPoolConfiguration;
            }

            @Override
            public DataSourceImplementation dataSourceImplementation() {
                return dataSourceImplementation;
            }

            @Override
            public boolean metricsEnabled() {
                return metrics;
            }

            @Override
            public void setMetricsEnabled(boolean metricsEnabled) {
                metrics = metricsEnabled;
                if (metricsEnabledListener != null) {
                    metricsEnabledListener.onMetricsEnabled(metricsEnabled);
                }
            }

            @Override
            public void registerMetricsEnabledListener(MetricsEnabledListener listener) {
                metricsEnabledListener = listener;
            }
        };
    }
}
