package org.xbib.jdbc.pool.api.configuration.supplier;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.nio.file.Path;
import java.time.Duration;
import java.util.Map;
import java.util.Properties;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Supplier;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionFactoryConfiguration.TransactionIsolation;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.ConnectionValidator;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.ExceptionSorter;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.MultipleAcquisitionAction;
import org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.TransactionRequirement;
import org.xbib.jdbc.pool.api.configuration.XbibDataSourceConfiguration;
import org.xbib.jdbc.pool.api.configuration.XbibDataSourceConfiguration.DataSourceImplementation;
import org.xbib.jdbc.pool.api.exceptionsorter.DB2ExceptionSorter;
import org.xbib.jdbc.pool.api.exceptionsorter.MSSQLExceptionSorter;
import org.xbib.jdbc.pool.api.exceptionsorter.MySQLExceptionSorter;
import org.xbib.jdbc.pool.api.exceptionsorter.OracleExceptionSorter;
import org.xbib.jdbc.pool.api.exceptionsorter.PostgreSQLExceptionSorter;
import org.xbib.jdbc.pool.api.exceptionsorter.SybaseExceptionSorter;
import org.xbib.jdbc.pool.api.security.NamePrincipal;
import org.xbib.jdbc.pool.api.security.SimplePassword;
import static java.lang.Long.parseLong;
import static java.util.function.Function.identity;
import static org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.ConnectionValidator.defaultValidator;
import static org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.ConnectionValidator.defaultValidatorWithTimeout;
import static org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.ConnectionValidator.emptyValidator;
import static org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.ExceptionSorter.defaultExceptionSorter;
import static org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.ExceptionSorter.emptyExceptionSorter;
import static org.xbib.jdbc.pool.api.configuration.XbibConnectionPoolConfiguration.ExceptionSorter.fatalExceptionSorter;

/**
 * Convenient way to build a configuration.
 * This class can build a configuration from a *.properties file or a {@link Properties} object.
 * This class defines keys for all the options and also allows for a prefix when looking for that properties.
 */
@SuppressWarnings("WeakerAccess")
public class XbibPropertiesReader implements Supplier<XbibDataSourceConfiguration> {

    public static final String IMPLEMENTATION = "implementation";
    public static final String METRICS_ENABLED = "metricsEnabled";

    public static final String MIN_SIZE = "minSize";
    public static final String MAX_SIZE = "maxSize";
    public static final String INITIAL_SIZE = "initialSize";
    public static final String FLUSH_ON_CLOSE = "flushOnClose";
    public static final String CONNECTION_VALIDATOR = "connectionValidator";
    public static final String ENHANCED_LEAK_REPORT = "enhancedLeakReport";
    public static final String EXCEPTION_SORTER = "exceptionSorter";
    public static final String MULTIPLE_ACQUISITION = "multipleAcquisition";
    public static final String TRANSACTION_REQUIREMENT = "transactionRequirement";
    public static final String VALIDATE_ON_BORROW = "validateOnBorrow";

    public static final String ACQUISITION_TIMEOUT = "acquisitionTimeout";
    public static final String ACQUISITION_TIMEOUT_MS = "acquisitionTimeout_ms";
    public static final String ACQUISITION_TIMEOUT_S = "acquisitionTimeout_s";
    public static final String ACQUISITION_TIMEOUT_M = "acquisitionTimeout_m";

    public static final String IDLE_VALIDATION = "idleValidation";
    public static final String IDLE_VALIDATION_MS = "idleValidation_ms";
    public static final String IDLE_VALIDATION_S = "idleValidation_s";
    public static final String IDLE_VALIDATION_M = "idleValidation_m";

    public static final String VALIDATION_TIMEOUT = "validationTimeout";
    public static final String VALIDATION_TIMEOUT_MS = "validationTimeout_ms";
    public static final String VALIDATION_TIMEOUT_S = "validationTimeout_s";
    public static final String VALIDATION_TIMEOUT_M = "validationTimeout_m";

    public static final String LEAK_TIMEOUT = "leakTimeout";
    public static final String LEAK_TIMEOUT_MS = "leakTimeout_ms";
    public static final String LEAK_TIMEOUT_S = "leakTimeout_s";
    public static final String LEAK_TIMEOUT_M = "leakTimeout_m";

    public static final String REAP_TIMEOUT = "reapTimeout";
    public static final String REAP_TIMEOUT_MS = "reapTimeout_ms";
    public static final String REAP_TIMEOUT_S = "reapTimeout_s";
    public static final String REAP_TIMEOUT_M = "reapTimeout_m";

    public static final String MAX_LIFETIME = "maxLifetime";
    public static final String MAX_LIFETIME_MS = "maxLifetime_ms";
    public static final String MAX_LIFETIME_S = "maxLifetime_s";
    public static final String MAX_LIFETIME_M = "maxLifetime_m";

    // --- //

    public static final String JDBC_URL = "jdbcUrl";
    public static final String AUTO_COMMIT = "autoCommit";
    public static final String TRACK_JDBC_RESOURCES = "trackJdbcResources";
    public static final String LOGIN_TIMEOUT = "loginTimeout";
    public static final String INITIAL_SQL = "initialSQL";
    public static final String PROVIDER_CLASS_NAME = "providerClassName";
    public static final String TRANSACTION_ISOLATION = "jdbcTransactionIsolation";
    public static final String PRINCIPAL = "principal";
    public static final String CREDENTIAL = "credential";
    public static final String RECOVERY_PRINCIPAL = "recoveryPrincipal";
    public static final String RECOVERY_CREDENTIAL = "recoveryCredential";
    public static final String JDBC_PROPERTIES = "jdbcProperties";
    public static final String XA_PROPERTIES = "xaProperties";

    // --- //

    private final String prefix;

    private final XbibDataSourceConfigurationSupplier dataSourceSupplier;

    public XbibPropertiesReader() {
        this("");
    }

    public XbibPropertiesReader(String readerPrefix) {
        prefix = readerPrefix;
        dataSourceSupplier = new XbibDataSourceConfigurationSupplier();
    }

    /**
     * Accepts the following options:
     * <ul>
     * <li>`empty` for the default {@link ConnectionValidator#emptyValidator()}</li>
     * <li>`default` for {@link ConnectionValidator#defaultValidator()}</li>
     * <li>`defaultX` for {@link ConnectionValidator#defaultValidatorWithTimeout(int)} where `X` is the timeout in seconds</li>
     * <li>the name of a class that implements {@link ConnectionValidator}</li>
     * </ul>
     */
    public static ConnectionValidator parseConnectionValidator(String connectionValidatorName) {
        if ("empty".equalsIgnoreCase(connectionValidatorName)) {
            return emptyValidator();
        } else if ("default".equalsIgnoreCase(connectionValidatorName)) {
            return defaultValidator();
        } else if (connectionValidatorName.regionMatches(true, 0, "default", 0, "default".length())) {
            return defaultValidatorWithTimeout((int) parseDurationS(connectionValidatorName.substring("default".length())).toSeconds());
        }

        try {
            Class<? extends ConnectionValidator> validatorClass = Thread.currentThread().getContextClassLoader().loadClass(connectionValidatorName).asSubclass(ConnectionValidator.class);
            return validatorClass.getDeclaredConstructor().newInstance();
        } catch (ClassNotFoundException e) {
            throw new IllegalArgumentException("Unknown connection validator " + connectionValidatorName);
        } catch (ClassCastException e) {
            throw new IllegalArgumentException(connectionValidatorName + " class is not a ConnectionValidator", e);
        } catch (InvocationTargetException | InstantiationException | IllegalAccessException |
                 NoSuchMethodException e) {
            throw new IllegalArgumentException("Unable to instantiate ConnectionValidator " + connectionValidatorName, e);
        }
    }

    /**
     * Accepts the following options:
     * <ul>
     * <li>`empty` for the default {@link ExceptionSorter#emptyExceptionSorter()}</li>
     * <li>`default` for {@link ExceptionSorter#defaultExceptionSorter()}</li>
     * <li>`fatal` for {@link ExceptionSorter#fatalExceptionSorter()}</li>
     * <li>`DB2` for the {@link DB2ExceptionSorter}</li>
     * <li>`MSSQL` for the {@link MSSQLExceptionSorter}</li>
     * <li>`MySQL` for the {@link MySQLExceptionSorter}</li>
     * <li>`Oracle` for the {@link OracleExceptionSorter}</li>
     * <li>`Postgres` or `PostgreSQL` for the {@link PostgreSQLExceptionSorter}</li>
     * <li>`Sybase` for the {@link SybaseExceptionSorter}</li>
     * <li>the name of a class that implements {@link ExceptionSorter}</li>
     * </ul>
     */
    public static ExceptionSorter parseExceptionSorter(String exceptionSorterName) {
        if ("empty".equalsIgnoreCase(exceptionSorterName)) {
            return emptyExceptionSorter();
        } else if ("default".equalsIgnoreCase(exceptionSorterName)) {
            return defaultExceptionSorter();
        } else if ("fatal".equalsIgnoreCase(exceptionSorterName)) {
            return fatalExceptionSorter();
        } else if ("DB2".equalsIgnoreCase(exceptionSorterName)) {
            return new DB2ExceptionSorter();
        } else if ("MSSQL".equalsIgnoreCase(exceptionSorterName)) {
            return new MSSQLExceptionSorter();
        } else if ("MySQL".equalsIgnoreCase(exceptionSorterName)) {
            return new MySQLExceptionSorter();
        } else if ("Oracle".equalsIgnoreCase(exceptionSorterName)) {
            return new OracleExceptionSorter();
        } else if ("Postgres".equalsIgnoreCase(exceptionSorterName) || "PostgreSQL".equalsIgnoreCase(exceptionSorterName)) {
            return new PostgreSQLExceptionSorter();
        } else if ("Sybase".equalsIgnoreCase(exceptionSorterName)) {
            return new SybaseExceptionSorter();
        }
        try {
            Class<? extends ExceptionSorter> sorterClass = Thread.currentThread().getContextClassLoader().loadClass(exceptionSorterName).asSubclass(ExceptionSorter.class);
            return sorterClass.getDeclaredConstructor().newInstance();
        } catch (ClassNotFoundException e) {
            throw new IllegalArgumentException("Unknown exception sorter " + exceptionSorterName);
        } catch (ClassCastException e) {
            throw new IllegalArgumentException(exceptionSorterName + " class is not a ExceptionSorter", e);
        } catch (InvocationTargetException | InstantiationException | IllegalAccessException |
                 NoSuchMethodException e) {
            throw new IllegalArgumentException("Unable to instantiate ExceptionSorter " + exceptionSorterName, e);
        }
    }

    // --- //

    private static Duration parseDurationMs(String value) {
        return Duration.ofMillis(parseLong(value));
    }

    private static Duration parseDurationS(String value) {
        return Duration.ofSeconds(parseLong(value));
    }

    private static Duration parseDurationM(String value) {
        return Duration.ofMinutes(parseLong(value));
    }

    @Override
    public XbibDataSourceConfiguration get() {
        return dataSourceSupplier.get();
    }

    public XbibDataSourceConfigurationSupplier modify() {
        return dataSourceSupplier;
    }

    public XbibPropertiesReader readProperties(Path path) throws IOException {
        return readProperties(path.toFile());
    }

    public XbibPropertiesReader readProperties(String filename) throws IOException {
        return readProperties(new File(filename));
    }

    // --- //

    public XbibPropertiesReader readProperties(File file) throws IOException {
        try (InputStream inputStream = new FileInputStream(file)) {
            Properties properties = new Properties();
            properties.load(inputStream);
            return readProperties(properties);
        }
    }

    // --- //

    @SuppressWarnings("unchecked")
    public XbibPropertiesReader readProperties(Properties properties) {
        return readProperties((Map) properties);
    }

    // --- //

    public XbibPropertiesReader readProperties(Map<String, String> properties) {
        XbibConnectionPoolConfigurationSupplier connectionPoolSupplier = new XbibConnectionPoolConfigurationSupplier();
        XbibConnectionFactoryConfigurationSupplier connectionFactorySupplier = new XbibConnectionFactoryConfigurationSupplier();

        apply(dataSourceSupplier::dataSourceImplementation, DataSourceImplementation::valueOf, properties, IMPLEMENTATION);
        apply(dataSourceSupplier::metricsEnabled, Boolean::parseBoolean, properties, METRICS_ENABLED);

        apply(connectionPoolSupplier::minSize, Integer::parseInt, properties, MIN_SIZE);
        apply(connectionPoolSupplier::maxSize, Integer::parseInt, properties, MAX_SIZE);
        apply(connectionPoolSupplier::flushOnClose, Boolean::parseBoolean, properties, FLUSH_ON_CLOSE);
        apply(connectionPoolSupplier::initialSize, Integer::parseInt, properties, INITIAL_SIZE);
        apply(connectionPoolSupplier::connectionValidator, XbibPropertiesReader::parseConnectionValidator, properties, CONNECTION_VALIDATOR);
        apply(connectionPoolSupplier::exceptionSorter, XbibPropertiesReader::parseExceptionSorter, properties, EXCEPTION_SORTER);
        apply(connectionPoolSupplier::enhancedLeakReport, Boolean::parseBoolean, properties, ENHANCED_LEAK_REPORT);
        apply(connectionPoolSupplier::multipleAcquisition, MultipleAcquisitionAction::valueOf, properties, MULTIPLE_ACQUISITION);
        apply(connectionPoolSupplier::transactionRequirement, TransactionRequirement::valueOf, properties, TRANSACTION_REQUIREMENT);
        apply(connectionPoolSupplier::validateOnBorrow, Boolean::parseBoolean, properties, VALIDATE_ON_BORROW);

        apply(connectionPoolSupplier::acquisitionTimeout, Duration::parse, properties, ACQUISITION_TIMEOUT);
        apply(connectionPoolSupplier::acquisitionTimeout, XbibPropertiesReader::parseDurationMs, properties, ACQUISITION_TIMEOUT_MS);
        apply(connectionPoolSupplier::acquisitionTimeout, XbibPropertiesReader::parseDurationS, properties, ACQUISITION_TIMEOUT_S);
        apply(connectionPoolSupplier::acquisitionTimeout, XbibPropertiesReader::parseDurationM, properties, ACQUISITION_TIMEOUT_M);

        apply(connectionPoolSupplier::idleValidationTimeout, Duration::parse, properties, IDLE_VALIDATION);
        apply(connectionPoolSupplier::idleValidationTimeout, XbibPropertiesReader::parseDurationMs, properties, IDLE_VALIDATION_MS);
        apply(connectionPoolSupplier::idleValidationTimeout, XbibPropertiesReader::parseDurationS, properties, IDLE_VALIDATION_S);
        apply(connectionPoolSupplier::idleValidationTimeout, XbibPropertiesReader::parseDurationM, properties, IDLE_VALIDATION_M);

        apply(connectionPoolSupplier::validationTimeout, Duration::parse, properties, VALIDATION_TIMEOUT);
        apply(connectionPoolSupplier::validationTimeout, XbibPropertiesReader::parseDurationMs, properties, VALIDATION_TIMEOUT_MS);
        apply(connectionPoolSupplier::validationTimeout, XbibPropertiesReader::parseDurationS, properties, VALIDATION_TIMEOUT_S);
        apply(connectionPoolSupplier::validationTimeout, XbibPropertiesReader::parseDurationM, properties, VALIDATION_TIMEOUT_M);

        apply(connectionPoolSupplier::leakTimeout, Duration::parse, properties, LEAK_TIMEOUT);
        apply(connectionPoolSupplier::leakTimeout, XbibPropertiesReader::parseDurationMs, properties, LEAK_TIMEOUT_MS);
        apply(connectionPoolSupplier::leakTimeout, XbibPropertiesReader::parseDurationS, properties, LEAK_TIMEOUT_S);
        apply(connectionPoolSupplier::leakTimeout, XbibPropertiesReader::parseDurationM, properties, LEAK_TIMEOUT_M);

        apply(connectionPoolSupplier::reapTimeout, Duration::parse, properties, REAP_TIMEOUT);
        apply(connectionPoolSupplier::reapTimeout, XbibPropertiesReader::parseDurationMs, properties, REAP_TIMEOUT_MS);
        apply(connectionPoolSupplier::reapTimeout, XbibPropertiesReader::parseDurationS, properties, REAP_TIMEOUT_S);
        apply(connectionPoolSupplier::reapTimeout, XbibPropertiesReader::parseDurationM, properties, REAP_TIMEOUT_M);

        apply(connectionPoolSupplier::maxLifetime, Duration::parse, properties, MAX_LIFETIME);
        apply(connectionPoolSupplier::maxLifetime, XbibPropertiesReader::parseDurationMs, properties, MAX_LIFETIME_MS);
        apply(connectionPoolSupplier::maxLifetime, XbibPropertiesReader::parseDurationS, properties, MAX_LIFETIME_S);
        apply(connectionPoolSupplier::maxLifetime, XbibPropertiesReader::parseDurationM, properties, MAX_LIFETIME_M);

        apply(connectionFactorySupplier::jdbcUrl, identity(), properties, JDBC_URL);
        apply(connectionFactorySupplier::autoCommit, Boolean::parseBoolean, properties, AUTO_COMMIT);
        apply(connectionFactorySupplier::trackJdbcResources, Boolean::parseBoolean, properties, TRACK_JDBC_RESOURCES);
        apply(connectionFactorySupplier::loginTimeout, Duration::parse, properties, LOGIN_TIMEOUT);
        apply(connectionFactorySupplier::initialSql, identity(), properties, INITIAL_SQL);
        apply(connectionFactorySupplier::connectionProviderClassName, identity(), properties, PROVIDER_CLASS_NAME);
        apply(connectionFactorySupplier::jdbcTransactionIsolation, TransactionIsolation::valueOf, properties, TRANSACTION_ISOLATION);
        apply(connectionFactorySupplier::principal, NamePrincipal::new, properties, PRINCIPAL);
        apply(connectionFactorySupplier::credential, SimplePassword::new, properties, CREDENTIAL);
        apply(connectionFactorySupplier::recoveryPrincipal, NamePrincipal::new, properties, RECOVERY_PRINCIPAL);
        apply(connectionFactorySupplier::recoveryCredential, SimplePassword::new, properties, RECOVERY_CREDENTIAL);
        applyJdbcProperties(connectionFactorySupplier::jdbcProperty, properties, JDBC_PROPERTIES);
        applyJdbcProperties(connectionFactorySupplier::xaProperty, properties, XA_PROPERTIES);

        dataSourceSupplier.connectionPoolConfiguration(connectionPoolSupplier.connectionFactoryConfiguration(connectionFactorySupplier));
        return this;
    }

    @SuppressWarnings("StringConcatenation")
    private <T> void apply(Consumer<? super T> consumer, Function<? super String, T> converter, Map<String, String> properties, String key) {
        String value = properties.get(prefix + key);
        if (value != null) {
            consumer.accept(converter.apply(value));
        }
    }

    @SuppressWarnings({"SameParameterValue", "StringConcatenation"})
    private void applyJdbcProperties(BiConsumer<? super String, ? super String> consumer, Map<String, String> properties, String key) {
        String propertiesArray = properties.get(prefix + key);
        if (propertiesArray != null && !propertiesArray.isEmpty()) {
            for (String property : propertiesArray.split(";")) {
                String[] keyValue = property.split("=");
                consumer.accept(keyValue[0], keyValue[1]);
            }
        }
    }
}
